@@ -55,7 +55,7 @@ gint at_util_call_compare_by_status(gconstpointer a, gconstpointer b);
 gint at_util_call_compare_by_phone_number(gconstpointer a, gconstpointer b);
 gint at_util_call_compare_by_id(gconstpointer a, gconstpointer b);
 gint at_util_call_compare(gconstpointer a, gconstpointer b);
-GSList *at_util_parse_clcc(GAtResult *result);
+GSList *at_util_parse_clcc(GAtResult *result, unsigned int *mpty_ids);
 gboolean at_util_parse_reg(GAtResult *result, const char *prefix,
 				int *mode, int *status,
 				int *lac, int *ci, int *tech,
@@ -112,11 +112,6 @@ static inline int at_util_convert_signal_strength(int strength)
 	return result;
 }
 
-#define DECLARE_FAILURE(e) 			\
-	struct ofono_error e;			\
-	e.type = OFONO_ERROR_TYPE_FAILURE;	\
-	e.error = 0				\
-
 #define CALLBACK_WITH_FAILURE(cb, args...)		\
 	do {						\
 		struct ofono_error cb_e;		\
