@@ -26,6 +26,8 @@
 #define _GNU_SOURCE
 #include <stdio.h>
 #include <ctype.h>
+#include <dirent.h>
+#include <errno.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -33,12 +35,18 @@
 
 #include <glib.h>
 
+#include <ofono/log.h>
+
 #include "qmi.h"
 #include "ctl.h"
 
 typedef void (*qmi_message_func_t)(uint16_t message, uint16_t length,
 					const void *buffer, void *user_data);
 
+struct discovery {
+	qmi_destroy_func_t destroy;
+};
+
 struct qmi_device {
 	int ref_count;
 	int fd;
@@ -49,6 +57,7 @@ struct qmi_device {
 	GQueue *req_queue;
 	GQueue *control_queue;
 	GQueue *service_queue;
+	GQueue *discovery_queue;
 	uint8_t next_control_tid;
 	uint16_t next_service_tid;
 	qmi_debug_func_t debug_func;
@@ -60,6 +69,10 @@ struct qmi_device {
 	uint8_t version_count;
 	GHashTable *service_list;
 	unsigned int release_users;
+	qmi_shutdown_func_t shutdown_func;
+	void *shutdown_user_data;
+	qmi_destroy_func_t shutdown_destroy;
+	guint shutdown_source;
 };
 
 struct qmi_service {
@@ -209,6 +222,14 @@ static gint __request_compare(gconstpointer a, gconstpointer b)
 	return req->tid - tid;
 }
 
+static void __discovery_free(gpointer data, gpointer user_data)
+{
+	struct discovery *d = data;
+	qmi_destroy_func_t destroy = d->destroy;
+
+	destroy(d);
+}
+
 static void __notify_free(gpointer data, gpointer user_data)
 {
 	struct qmi_notify *notify = data;
@@ -313,8 +334,12 @@ static const char *__service_type_to_string(uint8_t type)
 		return "UIM";
 	case QMI_SERVICE_PBM:
 		return "PBM";
+	case QMI_SERVICE_QCHAT:
+		return "QCHAT";
 	case QMI_SERVICE_RMTFS:
 		return "RMTFS";
+	case QMI_SERVICE_TEST:
+		return "TEST";
 	case QMI_SERVICE_LOC:
 		return "LOC";
 	case QMI_SERVICE_SAR:
@@ -326,7 +351,21 @@ static const char *__service_type_to_string(uint8_t type)
 	case QMI_SERVICE_TS:
 		return "TS";
 	case QMI_SERVICE_TMD:
-		return "TMS";
+		return "TMD";
+	case QMI_SERVICE_WDA:
+		return "WDA";
+	case QMI_SERVICE_CSVT:
+		return "CSVT";
+	case QMI_SERVICE_COEX:
+		return "COEX";
+	case QMI_SERVICE_PDC:
+		return "PDC";
+	case QMI_SERVICE_RFRPE:
+		return "RFRPE";
+	case QMI_SERVICE_DSD:
+		return "DSD";
+	case QMI_SERVICE_SSCTL:
+		return "SSCTL";
 	case QMI_SERVICE_CAT_OLD:
 		return "CAT";
 	case QMI_SERVICE_RMS:
@@ -338,6 +377,103 @@ static const char *__service_type_to_string(uint8_t type)
 	return NULL;
 }
 
+static const struct {
+	uint16_t err;
+	const char *str;
+} __error_table[] = {
+	{ 0x0000, "NONE"			},
+	{ 0x0001, "MALFORMED_MSG"		},
+	{ 0x0002, "NO_MEMORY"			},
+	{ 0x0003, "INTERNAL"			},
+	{ 0x0004, "ABORTED"			},
+	{ 0x0005, "CLIENT_IDS_EXHAUSTED"	},
+	{ 0x0006, "UNABORTABLE_TRANSACTION"	},
+	{ 0x0007, "INVALID_CLIENT_ID"		},
+	{ 0x0008, "NO_THRESHOLDS"		},
+	{ 0x0009, "INVALID_HANDLE"		},
+	{ 0x000a, "INVALID_PROFILE"		},
+	{ 0x000b, "INVALID_PINID"		},
+	{ 0x000c, "INCORRECT_PIN"		},
+	{ 0x000d, "NO_NETWORK_FOUND"		},
+	{ 0x000e, "CALL_FAILED"			},
+	{ 0x000f, "OUT_OF_CALL"			},
+	{ 0x0010, "NOT_PROVISIONED"		},
+	{ 0x0011, "MISSING_ARG"			},
+	{ 0x0013, "ARG_TOO_LONG"		},
+	{ 0x0016, "INVALID_TX_ID"		},
+	{ 0x0017, "DEVICE_IN_USE"		},
+	{ 0x0018, "OP_NETWORK_UNSUPPORTED"	},
+	{ 0x0019, "OP_DEVICE_UNSUPPORTED"	},
+	{ 0x001a, "NO_EFFECT"			},
+	{ 0x001b, "NO_FREE_PROFILE"		},
+	{ 0x001c, "INVALID_PDP_TYPE"		},
+	{ 0x001d, "INVALID_TECH_PREF"		},
+	{ 0x001e, "INVALID_PROFILE_TYPE"	},
+	{ 0x001f, "INVALID_SERVICE_TYPE"	},
+	{ 0x0020, "INVALID_REGISTER_ACTION"	},
+	{ 0x0021, "INVALID_PS_ATTACH_ACTION"	},
+	{ 0x0022, "AUTHENTICATION_FAILED"	},
+	{ 0x0023, "PIN_BLOCKED"			},
+	{ 0x0024, "PIN_PERM_BLOCKED"		},
+	{ 0x0025, "UIM_NOT_INITIALIZED"		},
+	{ 0x0026, "MAX_QOS_REQUESTS_IN_USE"	},
+	{ 0x0027, "INCORRECT_FLOW_FILTER"	},
+	{ 0x0028, "NETWORK_QOS_UNAWARE"		},
+	{ 0x0029, "INVALID_QOS_ID/INVALID_ID"	},
+	{ 0x002a, "REQUESTED_NUM_UNSUPPORTED"	},
+	{ 0x002b, "INTERFACE_NOT_FOUND"		},
+	{ 0x002c, "FLOW_SUSPENDED"		},
+	{ 0x002d, "INVALID_DATA_FORMAT"		},
+	{ 0x002e, "GENERAL"			},
+	{ 0x002f, "UNKNOWN"			},
+	{ 0x0030, "INVALID_ARG"			},
+	{ 0x0031, "INVALID_INDEX"		},
+	{ 0x0032, "NO_ENTRY"			},
+	{ 0x0033, "DEVICE_STORAGE_FULL"		},
+	{ 0x0034, "DEVICE_NOT_READY"		},
+	{ 0x0035, "NETWORK_NOT_READY"		},
+	{ 0x0036, "CAUSE_CODE"			},
+	{ 0x0037, "MESSAGE_NOT_SENT"		},
+	{ 0x0038, "MESSAGE_DELIVERY_FAILURE"	},
+	{ 0x0039, "INVALID_MESSAGE_ID"		},
+	{ 0x003a, "ENCODING"			},
+	{ 0x003b, "AUTHENTICATION_LOCK"		},
+	{ 0x003c, "INVALID_TRANSACTION"		},
+	{ 0x0041, "SESSION_INACTIVE"		},
+	{ 0x0042, "SESSION_INVALID"		},
+	{ 0x0043, "SESSION_OWNERSHIP"		},
+	{ 0x0044, "INSUFFICIENT_RESOURCES"	},
+	{ 0x0045, "DISABLED"			},
+	{ 0x0046, "INVALID_OPERATION"		},
+	{ 0x0047, "INVALID_QMI_CMD"		},
+	{ 0x0048, "TPDU_TYPE"			},
+	{ 0x0049, "SMSC_ADDR"			},
+	{ 0x004a, "INFO_UNAVAILABLE"		},
+	{ 0x004b, "SEGMENT_TOO_LONG"		},
+	{ 0x004c, "SEGEMENT_ORDER"		},
+	{ 0x004d, "BUNDLING_NOT_SUPPORTED"	},
+	{ 0x004f, "POLICY_MISMATCH"		},
+	{ 0x0050, "SIM_FILE_NOT_FOUND"		},
+	{ 0x0051, "EXTENDED_INTERNAL"		},
+	{ 0x0052, "ACCESS_DENIED"		},
+	{ 0x0053, "HARDWARE_RESTRICTED"		},
+	{ 0x0054, "ACK_NOT_SENT"		},
+	{ 0x0055, "INJECT_TIMEOUT"		},
+	{ }
+};
+
+static const char *__error_to_string(uint16_t error)
+{
+	int i;
+
+	for (i = 0; __error_table[i].str; i++) {
+		if (__error_table[i].err == error)
+			return __error_table[i].str;
+	}
+
+	return NULL;
+}
+
 static void __debug_msg(const char dir, const void *buf, size_t len,
 				qmi_debug_func_t function, void *user_data)
 {
@@ -347,6 +483,7 @@ static void __debug_msg(const char dir, const void *buf, size_t len,
 	const void *ptr;
 	uint16_t offset;
 	char strbuf[72 + 16], *str;
+	bool pending_print = false;
 
 	if (!function || !len)
 		return;
@@ -442,25 +579,36 @@ static void __debug_msg(const char dir, const void *buf, size_t len,
 		if (tlv->type == 0x02 && tlv_length == QMI_RESULT_CODE_SIZE) {
 			const struct qmi_result_code *result = ptr + offset +
 							QMI_TLV_HDR_SIZE;
+			uint16_t error = GUINT16_FROM_LE(result->error);
+			const char *error_str;
 
-			str += sprintf(str, " {type=%d,error=%d}", tlv->type,
-						GUINT16_FROM_LE(result->error));
+			error_str = __error_to_string(error);
+			if (error_str)
+				str += sprintf(str, " {type=%d,error=%s}",
+							tlv->type, error_str);
+			else
+				str += sprintf(str, " {type=%d,error=%d}",
+							tlv->type, error);
 		} else {
 			str += sprintf(str, " {type=%d,len=%d}", tlv->type,
 								tlv_length);
 		}
 
-		if (str - strbuf > 72) {
+		if (str - strbuf > 60) {
 			function(strbuf, user_data);
 
 			str = strbuf;
 			str += sprintf(str, "      ");
-		}
+
+			pending_print = false;
+		} else
+			pending_print = true;
 
 		offset += QMI_TLV_HDR_SIZE + tlv_length;
 	}
 
-	function(strbuf, user_data);
+	if (pending_print)
+		function(strbuf, user_data);
 }
 
 static void __debug_device(struct qmi_device *device,
@@ -647,7 +795,7 @@ static void handle_packet(struct qmi_device *device,
 
 		tid = GUINT16_FROM_LE(service->transaction);
 
-		if (service->type == 0x04 && tid == 0x0000) {
+		if (service->type == 0x04) {
 			handle_indication(device, hdr->service, hdr->client,
 							message, length, data);
 			return;
@@ -727,6 +875,21 @@ static void read_watch_destroy(gpointer user_data)
 	device->read_watch = 0;
 }
 
+static void __qmi_device_discovery_started(struct qmi_device *device,
+						struct discovery *d)
+{
+	g_queue_push_tail(device->discovery_queue, d);
+}
+
+static void __qmi_device_discovery_complete(struct qmi_device *device,
+						struct discovery *d)
+{
+	if (g_queue_remove(device->discovery_queue, d) != TRUE)
+		return;
+
+	__discovery_free(d, NULL);
+}
+
 static void service_destroy(gpointer data)
 {
 	struct qmi_service *service = data;
@@ -780,6 +943,7 @@ struct qmi_device *qmi_device_new(int fd)
 	device->req_queue = g_queue_new();
 	device->control_queue = g_queue_new();
 	device->service_queue = g_queue_new();
+	device->discovery_queue = g_queue_new();
 
 	device->service_list = g_hash_table_new_full(g_direct_hash,
 					g_direct_equal, NULL, service_destroy);
@@ -816,6 +980,9 @@ void qmi_device_unref(struct qmi_device *device)
 	g_queue_foreach(device->req_queue, __request_free, NULL);
 	g_queue_free(device->req_queue);
 
+	g_queue_foreach(device->discovery_queue, __discovery_free, NULL);
+	g_queue_free(device->discovery_queue);
+
 	if (device->write_watch > 0)
 		g_source_remove(device->write_watch);
 
@@ -825,6 +992,9 @@ void qmi_device_unref(struct qmi_device *device)
 	if (device->close_on_unref)
 		close(device->fd);
 
+	if (device->shutdown_source)
+		g_source_remove(device->shutdown_source);
+
 	g_hash_table_destroy(device->service_list);
 
 	g_free(device->version_str);
@@ -876,6 +1046,7 @@ static const void *tlv_get(const void *data, uint16_t size,
 }
 
 struct discover_data {
+	struct discovery super;
 	struct qmi_device *device;
 	qmi_discover_func_t func;
 	void *user_data;
@@ -883,6 +1054,21 @@ struct discover_data {
 	guint timeout;
 };
 
+static void discover_data_free(gpointer user_data)
+{
+	struct discover_data *data = user_data;
+
+	if (data->timeout) {
+		g_source_remove(data->timeout);
+		data->timeout = 0;
+	}
+
+	if (data->destroy)
+		data->destroy(data->user_data);
+
+	g_free(data);
+}
+
 static void discover_callback(uint16_t message, uint16_t length,
 					const void *buffer, void *user_data)
 {
@@ -896,8 +1082,6 @@ static void discover_callback(uint16_t message, uint16_t length,
 	uint8_t count;
 	unsigned int i;
 
-	g_source_remove(data->timeout);
-
 	count = 0;
 	list = NULL;
 
@@ -925,6 +1109,7 @@ static void discover_callback(uint16_t message, uint16_t length,
 		uint16_t minor =
 			GUINT16_FROM_LE(service_list->services[i].minor);
 		uint8_t type = service_list->services[i].type;
+		const char *name = __service_type_to_string(type);
 
 		if (type == QMI_SERVICE_CONTROL) {
 			device->control_major = major;
@@ -935,9 +1120,16 @@ static void discover_callback(uint16_t message, uint16_t length,
 		list[count].type = type;
 		list[count].major = major;
 		list[count].minor = minor;
-		list[count].name = __service_type_to_string(type);
+		list[count].name = name;
 
 		count++;
+
+		if (name)
+			__debug_device(device, "found service [%s %d.%d]",
+							name, major, minor);
+		else
+			__debug_device(device, "found service [%d %d.%d]",
+							type, major, minor);
 	}
 
 	ptr = tlv_get(buffer, length, 0x10, &len);
@@ -960,10 +1152,7 @@ done:
 	if (data->func)
 		data->func(count, list, data->user_data);
 
-	if (data->destroy)
-		data->destroy(data->user_data);
-
-	g_free(data);
+	__qmi_device_discovery_complete(data->device, &data->super);
 }
 
 static gboolean discover_reply(gpointer user_data)
@@ -977,10 +1166,7 @@ static gboolean discover_reply(gpointer user_data)
 		data->func(device->version_count,
 				device->version_list, data->user_data);
 
-	if (data->destroy)
-		data->destroy(data->user_data);
-
-	g_free(data);
+	__qmi_device_discovery_complete(data->device, &data->super);
 
 	return FALSE;
 }
@@ -1001,13 +1187,15 @@ bool qmi_device_discover(struct qmi_device *device, qmi_discover_func_t func,
 	if (!data)
 		return false;
 
+	data->super.destroy = discover_data_free;
 	data->device = device;
 	data->func = func;
 	data->user_data = user_data;
 	data->destroy = destroy;
 
 	if (device->version_list) {
-		g_timeout_add_seconds(0, discover_reply, data);
+		data->timeout = g_timeout_add_seconds(0, discover_reply, data);
+		__qmi_device_discovery_started(device, &data->super);
 		return true;
 	}
 
@@ -1028,6 +1216,7 @@ bool qmi_device_discover(struct qmi_device *device, qmi_discover_func_t func,
 	__request_submit(device, req, hdr->transaction);
 
 	data->timeout = g_timeout_add_seconds(5, discover_reply, data);
+	__qmi_device_discovery_started(device, &data->super);
 
 	return true;
 }
@@ -1058,63 +1247,249 @@ static void release_client(struct qmi_device *device,
 	__request_submit(device, req, hdr->transaction);
 }
 
-struct shutdown_data {
-	struct qmi_device *device;
-	qmi_shutdown_func_t func;
-	void *user_data;
-	qmi_destroy_func_t destroy;
-};
-
-static gboolean shutdown_reply(gpointer user_data)
+static void shutdown_destroy(gpointer user_data)
 {
-	struct shutdown_data *data = user_data;
+	struct qmi_device *device = user_data;
 
-	if (data->func)
-		data->func(data->user_data);
+	if (device->shutdown_destroy)
+		device->shutdown_destroy(device->shutdown_user_data);
 
-	g_free(data);
-
-	return FALSE;
+	device->shutdown_source = 0;
 }
 
-static gboolean shutdown_timeout(gpointer user_data)
+static gboolean shutdown_callback(gpointer user_data)
 {
-	struct shutdown_data *data = user_data;
-	struct qmi_device *device = data->device;
+	struct qmi_device *device = user_data;
 
 	if (device->release_users > 0)
 		return TRUE;
 
-	return shutdown_reply(data);
+	if (device->shutdown_func)
+		device->shutdown_func(device->shutdown_user_data);
+
+	return FALSE;
 }
 
 bool qmi_device_shutdown(struct qmi_device *device, qmi_shutdown_func_t func,
 				void *user_data, qmi_destroy_func_t destroy)
 {
-	struct shutdown_data *data;
-
 	if (!device)
 		return false;
 
+	if (device->shutdown_source > 0)
+		return false;
+
 	__debug_device(device, "device %p shutdown", device);
 
-	data = g_try_new0(struct shutdown_data, 1);
-	if (!data)
+	device->shutdown_source = g_timeout_add_seconds_full(G_PRIORITY_DEFAULT,
+						0, shutdown_callback, device,
+						shutdown_destroy);
+	if (device->shutdown_source == 0)
 		return false;
 
-	data->device = device;
-	data->func = func;
-	data->user_data = user_data;
-	data->destroy = destroy;
-
-	if (device->release_users > 0)
-		g_timeout_add_seconds(0, shutdown_timeout, data);
-	else
-		g_timeout_add_seconds(0, shutdown_reply, data);
+	device->shutdown_func = func;
+	device->shutdown_user_data = user_data;
+	device->shutdown_destroy = destroy;
 
 	return true;
 }
 
+static bool get_device_file_name(struct qmi_device *device,
+					char *file_name, int size)
+{
+	pid_t pid;
+	char temp[100];
+	ssize_t result;
+
+	if (size <= 0)
+		return false;
+
+	pid = getpid();
+
+	snprintf(temp, 100, "/proc/%d/fd/%d", (int) pid, device->fd);
+	temp[99] = 0;
+
+	result = readlink(temp, file_name, size - 1);
+
+	if (result == -1 || result >= size - 1) {
+		DBG("Error %d in readlink", errno);
+		return false;
+	}
+
+	file_name[result] = 0;
+
+	return true;
+}
+
+static char *get_first_dir_in_directory(char *dir_path)
+{
+	DIR *dir;
+	struct dirent *dir_entry;
+	char *dir_name = NULL;
+
+	dir = opendir(dir_path);
+
+	if (!dir)
+		return NULL;
+
+	dir_entry = readdir(dir);
+
+	while ((dir_entry != NULL)) {
+		if (dir_entry->d_type == DT_DIR &&
+				strcmp(dir_entry->d_name, ".") != 0 &&
+				strcmp(dir_entry->d_name, "..") != 0) {
+			dir_name = g_strdup(dir_entry->d_name);
+			break;
+		}
+
+		dir_entry = readdir(dir);
+	}
+
+	closedir(dir);
+	return dir_name;
+}
+
+static char *get_device_interface(struct qmi_device *device)
+{
+	char * const driver_names[] = { "usbmisc", "usb" };
+	unsigned int i;
+	char file_path[PATH_MAX];
+	char *file_name;
+	char *interface = NULL;
+
+	if (!get_device_file_name(device, file_path, sizeof(file_path)))
+		return NULL;
+
+	file_name = basename(file_path);
+
+	for (i = 0; i < G_N_ELEMENTS(driver_names) && !interface; i++) {
+		gchar *sysfs_path;
+
+		sysfs_path = g_strdup_printf("/sys/class/%s/%s/device/net/",
+						driver_names[i], file_name);
+		interface = get_first_dir_in_directory(sysfs_path);
+		g_free(sysfs_path);
+	}
+
+	return interface;
+}
+
+enum qmi_device_expected_data_format qmi_device_get_expected_data_format(
+						struct qmi_device *device)
+{
+	char *sysfs_path = NULL;
+	char *interface = NULL;
+	int fd = -1;
+	char value;
+	enum qmi_device_expected_data_format expected =
+					QMI_DEVICE_EXPECTED_DATA_FORMAT_UNKNOWN;
+
+	if (!device)
+		goto done;
+
+	interface = get_device_interface(device);
+
+	if (!interface) {
+		DBG("Error while getting interface name");
+		goto done;
+	}
+
+	/* Build sysfs file path and open it */
+	sysfs_path = g_strdup_printf("/sys/class/net/%s/qmi/raw_ip", interface);
+
+	fd = open(sysfs_path, O_RDONLY);
+	if (fd < 0) {
+		/* maybe not supported by kernel */
+		DBG("Error %d in open(%s)", errno, sysfs_path);
+		goto done;
+	}
+
+	if (read(fd, &value, 1) != 1) {
+		DBG("Error %d in read(%s)", errno, sysfs_path);
+		goto done;
+	}
+
+	if (value == 'Y')
+		expected = QMI_DEVICE_EXPECTED_DATA_FORMAT_RAW_IP;
+	else if (value == 'N')
+		expected = QMI_DEVICE_EXPECTED_DATA_FORMAT_802_3;
+	else
+		DBG("Unexpected sysfs file contents");
+
+done:
+	if (fd >= 0)
+		close(fd);
+
+	if (sysfs_path)
+		g_free(sysfs_path);
+
+	if (interface)
+		g_free(interface);
+
+	return expected;
+}
+
+bool qmi_device_set_expected_data_format(struct qmi_device *device,
+			enum qmi_device_expected_data_format format)
+{
+	bool res = false;
+	char *sysfs_path = NULL;
+	char *interface = NULL;
+	int fd = -1;
+	char value;
+
+	if (!device)
+		goto done;
+
+	switch (format) {
+	case QMI_DEVICE_EXPECTED_DATA_FORMAT_802_3:
+		value = 'N';
+		break;
+	case QMI_DEVICE_EXPECTED_DATA_FORMAT_RAW_IP:
+		value = 'Y';
+		break;
+	default:
+		DBG("Unhandled format: %d", (int) format);
+		goto done;
+	}
+
+	interface = get_device_interface(device);
+
+	if (!interface) {
+		DBG("Error while getting interface name");
+		goto done;
+	}
+
+	/* Build sysfs file path and open it */
+	sysfs_path = g_strdup_printf("/sys/class/net/%s/qmi/raw_ip", interface);
+
+	fd = open(sysfs_path, O_WRONLY);
+	if (fd < 0) {
+		/* maybe not supported by kernel */
+		DBG("Error %d in open(%s)", errno, sysfs_path);
+		goto done;
+	}
+
+	if (write(fd, &value, 1) != 1) {
+		DBG("Error %d in write(%s)", errno, sysfs_path);
+		goto done;
+	}
+
+	res = true;
+
+done:
+	if (fd >= 0)
+		close(fd);
+
+	if (sysfs_path)
+		g_free(sysfs_path);
+
+	if (interface)
+		g_free(interface);
+
+	return res;
+}
+
 struct qmi_param *qmi_param_new(void)
 {
 	struct qmi_param *param;
@@ -1262,6 +1637,17 @@ bool qmi_result_set_error(struct qmi_result *result, uint16_t *error)
 	return true;
 }
 
+const char *qmi_result_get_error(struct qmi_result *result)
+{
+	if (!result)
+		return NULL;
+
+	if (result->result == 0x0000)
+		return NULL;
+
+	return __error_to_string(result->error);
+}
+
 const void *qmi_result_get(struct qmi_result *result, uint8_t type,
 							uint16_t *length)
 {
@@ -1371,6 +1757,7 @@ bool qmi_result_get_uint64(struct qmi_result *result, uint8_t type,
 }
 
 struct service_create_data {
+	struct discovery super;
 	struct qmi_device *device;
 	bool shared;
 	uint8_t type;
@@ -1382,16 +1769,29 @@ struct service_create_data {
 	guint timeout;
 };
 
-static gboolean service_create_reply(gpointer user_data)
+static void service_create_data_free(gpointer user_data)
 {
 	struct service_create_data *data = user_data;
 
-	data->func(NULL, data->user_data);
+	if (data->timeout) {
+		g_source_remove(data->timeout);
+		data->timeout = 0;
+	}
 
 	if (data->destroy)
 		data->destroy(data->user_data);
 
 	g_free(data);
+}
+
+static gboolean service_create_reply(gpointer user_data)
+{
+	struct service_create_data *data = user_data;
+
+	data->timeout = 0;
+	data->func(NULL, data->user_data);
+
+	__qmi_device_discovery_complete(data->device, &data->super);
 
 	return FALSE;
 }
@@ -1407,8 +1807,6 @@ static void service_create_callback(uint16_t message, uint16_t length,
 	uint16_t len;
 	unsigned int hash_id;
 
-	g_source_remove(data->timeout);
-
 	result_code = tlv_get(buffer, length, 0x02, &len);
 	if (!result_code)
 		goto done;
@@ -1450,13 +1848,9 @@ static void service_create_callback(uint16_t message, uint16_t length,
 
 done:
 	data->func(service, data->user_data);
-
 	qmi_service_unref(service);
 
-	if (data->destroy)
-		data->destroy(data->user_data);
-
-	g_free(data);
+	__qmi_device_discovery_complete(data->device, &data->super);
 }
 
 static void service_create_discover(uint8_t count,
@@ -1487,7 +1881,9 @@ static void service_create_discover(uint8_t count,
 		if (data->timeout > 0)
 			g_source_remove(data->timeout);
 
-		g_timeout_add_seconds(0, service_create_reply, data);
+		data->timeout = g_timeout_add_seconds(0,
+						service_create_reply, data);
+		__qmi_device_discovery_started(device, &data->super);
 		return;
 	}
 
@@ -1510,6 +1906,7 @@ static bool service_create(struct qmi_device *device, bool shared,
 	if (!data)
 		return false;
 
+	data->super.destroy = service_create_data_free;
 	data->device = device;
 	data->shared = shared;
 	data->type = type;
@@ -1532,6 +1929,7 @@ static bool service_create(struct qmi_device *device, bool shared,
 
 done:
 	data->timeout = g_timeout_add_seconds(8, service_create_reply, data);
+	__qmi_device_discovery_started(device, &data->super);
 
 	return true;
 }
@@ -1550,17 +1948,23 @@ bool qmi_service_create(struct qmi_device *device,
 }
 
 struct service_create_shared_data {
+	struct discovery super;
 	struct qmi_service *service;
+	struct qmi_device *device;
 	qmi_create_func_t func;
 	void *user_data;
 	qmi_destroy_func_t destroy;
+	guint timeout;
 };
 
-static gboolean service_create_shared_reply(gpointer user_data)
+static void service_create_shared_data_free(gpointer user_data)
 {
 	struct service_create_shared_data *data = user_data;
 
-	data->func(data->service, data->user_data);
+	if (data->timeout) {
+		g_source_remove(data->timeout);
+		data->timeout = 0;
+	}
 
 	qmi_service_unref(data->service);
 
@@ -1568,6 +1972,16 @@ static gboolean service_create_shared_reply(gpointer user_data)
 		data->destroy(data->user_data);
 
 	g_free(data);
+}
+
+static gboolean service_create_shared_reply(gpointer user_data)
+{
+	struct service_create_shared_data *data = user_data;
+
+	data->timeout = 0;
+	data->func(data->service, data->user_data);
+
+	__qmi_device_discovery_complete(data->device, &data->super);
 
 	return FALSE;
 }
@@ -1594,13 +2008,16 @@ bool qmi_service_create_shared(struct qmi_device *device,
 		if (!data)
 			return false;
 
+		data->super.destroy = service_create_shared_data_free;
 		data->service = qmi_service_ref(service);
-
+		data->device = device;
 		data->func = func;
 		data->user_data = user_data;
 		data->destroy = destroy;
 
-		g_timeout_add(0, service_create_shared_reply, data);
+		data->timeout = g_timeout_add(0,
+					service_create_shared_reply, data);
+		__qmi_device_discovery_started(device, &data->super);
 
 		return 0;
 	}
