@@ -62,6 +62,7 @@ static const char *gcap_prefix[] = { "+GCAP:", NULL };
 static const char *rfswitch_prefix[] = { "^RFSWITCH:", NULL };
 static const char *sysinfo_prefix[] = { "^SYSINFO:", NULL };
 static const char *ussdmode_prefix[] = { "^USSDMODE:", NULL };
+static const char *dialmode_prefix[] = { "^DIALMODE:", NULL };
 static const char *cvoice_prefix[] = { "^CVOICE:", NULL };
 
 enum {
@@ -77,6 +78,8 @@ enum {
 struct huawei_data {
 	GAtChat *modem;
 	GAtChat *pcui;
+	struct ofono_gprs *gprs;
+	struct ofono_gprs_context *gc;
 	gboolean have_sim;
 	int sim_state;
 	guint sysinfo_poll_source;
@@ -86,6 +89,8 @@ struct huawei_data {
 	gboolean have_voice;
 	gboolean have_gsm;
 	gboolean have_cdma;
+	gboolean have_ndis;
+	gboolean have_ussdmode;
 };
 
 static int huawei_probe(struct ofono_modem *modem)
@@ -146,12 +151,8 @@ static void ussdmode_query_cb(gboolean ok, GAtResult *result,
 	if (!g_at_result_iter_next_number(&iter, &ussdmode))
 		return;
 
-	if (ussdmode == 0)
-		return;
-
-	/* Set USSD mode to text mode */
-	g_at_chat_send(data->pcui, "AT^USSDMODE=0", none_prefix,
-						NULL, NULL, NULL);
+	if (ussdmode == 1)
+		data->have_ussdmode = TRUE;
 }
 
 static void ussdmode_support_cb(gboolean ok, GAtResult *result,
@@ -173,6 +174,77 @@ static void ussdmode_support_cb(gboolean ok, GAtResult *result,
 					ussdmode_query_cb, data, NULL);
 }
 
+static void dialmode_query_cb(gboolean ok, GAtResult *result,
+						gpointer user_data)
+{
+	struct huawei_data *data = user_data;
+	GAtResultIter iter;
+	gint dialmode, cdc_spec;
+	const char *str = "unknown";
+
+	if (!ok)
+		return;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^DIALMODE:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &dialmode))
+		return;
+
+	if (g_at_result_iter_next_number(&iter, &cdc_spec)) {
+		switch (cdc_spec) {
+		case 0:
+			str = "none";
+			break;
+		case 1:
+			str = "Modem port";
+			break;
+		case 2:
+			str = "NDIS port";
+			break;
+		case 3:
+			str = "Modem and NDIS port";
+			break;
+		}
+	}
+
+	switch (dialmode) {
+	case 0:
+		ofono_info("Modem support (CDC support: %s)", str);
+		data->have_ndis = FALSE;
+		break;
+	case 1:
+		ofono_info("NDIS support (CDC support: %s)", str);
+		data->have_ndis = TRUE;
+		break;
+	case 2:
+		ofono_info("Modem and NDIS support (CDC support: %s)", str);
+		data->have_ndis = TRUE;
+		break;
+	}
+}
+
+static void dialmode_support_cb(gboolean ok, GAtResult *result,
+						gpointer user_data)
+{
+	struct huawei_data *data = user_data;
+	GAtResultIter iter;
+
+	if (!ok)
+		return;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^DIALMODE:"))
+		return;
+
+	/* Query current NDIS mode */
+	g_at_chat_send(data->pcui, "AT^DIALMODE?", dialmode_prefix,
+					dialmode_query_cb, data, NULL);
+}
+
 static void cvoice_query_cb(gboolean ok, GAtResult *result,
 						gpointer user_data)
 {
@@ -347,6 +419,16 @@ static void sysinfo_enable_cb(gboolean ok, GAtResult *result,
 	g_at_chat_send(data->modem, "AT&C0", NULL, NULL, NULL, NULL);
 	g_at_chat_send(data->pcui, "AT&C0", NULL, NULL, NULL, NULL);
 
+	/*
+	 * Ensure that the modem is using GSM character set and not IRA,
+	 * otherwise weirdness with umlauts and other non-ASCII characters
+	 * can result
+	 */
+	g_at_chat_send(data->modem, "AT+CSCS=\"GSM\"", none_prefix,
+							NULL, NULL, NULL);
+	g_at_chat_send(data->pcui, "AT+CSCS=\"GSM\"", none_prefix,
+							NULL, NULL, NULL);
+
 	/* Query current device settings */
 	g_at_chat_send(data->pcui, "AT^U2DIAG?", none_prefix,
 						NULL, NULL, NULL);
@@ -359,10 +441,21 @@ static void sysinfo_enable_cb(gboolean ok, GAtResult *result,
 	g_at_chat_send(data->pcui, "AT^USSDMODE=?", ussdmode_prefix,
 					ussdmode_support_cb, data, NULL);
 
+	/* Check NDIS mode support */
+	g_at_chat_send(data->pcui, "AT^DIALMODE=?", dialmode_prefix,
+					dialmode_support_cb, data, NULL);
+
 	/* Check for voice support */
 	g_at_chat_send(data->pcui, "AT^CVOICE=?", cvoice_prefix,
 					cvoice_support_cb, modem, NULL);
 
+	/* For CDMA we use AlwaysOnline so we leave the modem online. */
+	if (data->have_gsm == FALSE && data->have_cdma == TRUE) {
+		ofono_modem_set_boolean(modem, "AlwaysOnline", TRUE);
+		ofono_modem_set_powered(modem, TRUE);
+		return;
+	}
+
 	if (g_at_chat_send(data->pcui, data->offline_command, none_prefix,
 					cfun_offline, modem, NULL) > 0)
 		return;
@@ -493,6 +586,48 @@ static GAtChat *open_device(struct ofono_modem *modem,
 	return chat;
 }
 
+static void modem_disconnect(gpointer user_data)
+{
+	struct ofono_modem *modem = user_data;
+	struct huawei_data *data = ofono_modem_get_data(modem);
+
+	if (data == NULL) {
+		DBG("Modem has already been removed");
+		return;
+	}
+
+	ofono_warn("Modem channel disconnected");
+
+	/* clean and close modem device */
+	g_at_chat_cancel_all(data->modem);
+	g_at_chat_unregister_all(data->modem);
+	g_at_chat_unref(data->modem);
+	data->modem = NULL;
+
+	/* close gprs context driver */
+	ofono_gprs_context_remove(data->gc);
+
+	/* reopen modem channel */
+	data->modem = open_device(modem, "Modem", "Modem: ");
+
+	if (data->modem == NULL) {
+		DBG("Can't reopen device");
+		return;
+	}
+
+	/* configure modem channel */
+	g_at_chat_set_disconnect_function(data->modem, modem_disconnect, modem);
+	g_at_chat_set_slave(data->modem, data->pcui);
+	g_at_chat_send(data->modem, "ATE0 +CMEE=1", NULL, NULL, NULL, NULL);
+
+	/* reopen gprs context driver */
+	data->gc = ofono_gprs_context_create(modem, OFONO_VENDOR_HUAWEI,
+						"atmodem", data->modem);
+
+	if (data->gprs && data->gc)
+		ofono_gprs_add_context(data->gprs, data->gc);
+}
+
 static int huawei_enable(struct ofono_modem *modem)
 {
 	struct huawei_data *data = ofono_modem_get_data(modem);
@@ -503,6 +638,8 @@ static int huawei_enable(struct ofono_modem *modem)
 	if (data->modem == NULL)
 		return -EINVAL;
 
+	g_at_chat_set_disconnect_function(data->modem, modem_disconnect, modem);
+
 	data->pcui = open_device(modem, "Pcui", "PCUI: ");
 	if (data->pcui == NULL) {
 		g_at_chat_unref(data->modem);
@@ -695,21 +832,25 @@ static void huawei_set_online(struct ofono_modem *modem, ofono_bool_t online,
 static void huawei_pre_sim(struct ofono_modem *modem)
 {
 	struct huawei_data *data = ofono_modem_get_data(modem);
+	struct ofono_sim *sim = NULL;
 
 	DBG("%p", modem);
 
 	if (data->have_gsm == TRUE) {
-		struct ofono_sim *sim;
-
 		ofono_devinfo_create(modem, 0, "atmodem", data->pcui);
 		sim = ofono_sim_create(modem, OFONO_VENDOR_HUAWEI,
 						"atmodem", data->pcui);
-
-		if (sim && data->have_sim == TRUE)
-			ofono_sim_inserted_notify(sim, TRUE);
 	} else if (data->have_cdma == TRUE) {
 		ofono_devinfo_create(modem, 0, "cdmamodem", data->pcui);
+
+		/* Create SIM atom only if SIM is not embedded */
+		if (data->sim_state != SIM_STATE_ROMSIM)
+			sim = ofono_sim_create(modem, OFONO_VENDOR_HUAWEI,
+						"atmodem-noef", data->pcui);
 	}
+
+	if (sim && data->have_sim == TRUE)
+		ofono_sim_inserted_notify(sim, TRUE);
 }
 
 static void huawei_post_sim(struct ofono_modem *modem)
@@ -725,9 +866,6 @@ static void huawei_post_sim(struct ofono_modem *modem)
 	}
 
 	if (data->have_gsm == TRUE) {
-		struct ofono_gprs *gprs;
-		struct ofono_gprs_context *gc;
-
 		ofono_phonebook_create(modem, 0, "atmodem", data->pcui);
 		ofono_radio_settings_create(modem, 0,
 						"huaweimodem", data->pcui);
@@ -735,13 +873,13 @@ static void huawei_post_sim(struct ofono_modem *modem)
 		ofono_sms_create(modem, OFONO_VENDOR_HUAWEI,
 						"atmodem", data->pcui);
 
-		gprs = ofono_gprs_create(modem, OFONO_VENDOR_HUAWEI,
+		data->gprs = ofono_gprs_create(modem, OFONO_VENDOR_HUAWEI,
 						"atmodem", data->pcui);
-		gc = ofono_gprs_context_create(modem, 0,
+		data->gc = ofono_gprs_context_create(modem, 0,
 						"atmodem", data->modem);
 
-		if (gprs && gc)
-			ofono_gprs_add_context(gprs, gc);
+		if (data->gprs && data->gc)
+			ofono_gprs_add_context(data->gprs, data->gc);
 	}
 }
 
@@ -757,8 +895,11 @@ static void huawei_post_online(struct ofono_modem *modem)
 
 		ofono_cbs_create(modem, OFONO_VENDOR_QUALCOMM_MSM,
 						"atmodem", data->pcui);
-		ofono_ussd_create(modem, OFONO_VENDOR_QUALCOMM_MSM,
-						"atmodem", data->pcui);
+		if (data->have_ussdmode == TRUE)
+			ofono_ussd_create(modem, 0, "huaweimodem", data->pcui);
+		else
+			ofono_ussd_create(modem, OFONO_VENDOR_QUALCOMM_MSM,
+							"atmodem", data->pcui);
 	} else if (data->have_cdma == TRUE) {
 		ofono_cdma_netreg_create(modem, 0, "huaweimodem", data->pcui);
 
