@@ -102,8 +102,10 @@ static struct ssc_entry *ssc_entry_create(const char *sc, void *cb, void *data,
 	return r;
 }
 
-static void ssc_entry_destroy(struct ssc_entry *ca)
+static void ssc_entry_destroy(gpointer pointer)
 {
+	struct ssc_entry *ca = pointer;
+
 	if (ca->destroy)
 		ca->destroy(ca->user);
 
@@ -414,6 +416,16 @@ void ofono_ussd_notify(struct ofono_ussd *ussd, int status, int dcs,
 		return;
 	}
 
+	if (status == OFONO_USSD_STATUS_TERMINATED) {
+		ussd_change_state(ussd, USSD_STATE_IDLE);
+
+		if (ussd->pending == NULL)
+			return;
+
+		reply = __ofono_error_network_terminated(ussd->pending);
+		goto out;
+	}
+
 	if (status == OFONO_USSD_STATUS_NOT_SUPPORTED) {
 		ussd_change_state(ussd, USSD_STATE_IDLE);
 
@@ -573,7 +585,7 @@ static DBusMessage *ussd_initiate(DBusConnection *conn, DBusMessage *msg,
 
 	DBG("No.., checking if this is a USSD string");
 	if (!valid_ussd_string(str, call_in_progress))
-		return __ofono_error_invalid_format(msg);
+		return __ofono_error_not_recognized(msg);
 
 	if (!ussd_encode(str, &num_packed, buf))
 		return __ofono_error_invalid_format(msg);
@@ -780,12 +792,10 @@ static void ussd_unregister(struct ofono_atom *atom)
 	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
 	const char *path = __ofono_atom_get_path(atom);
 
-	g_slist_foreach(ussd->ss_control_list, (GFunc) ssc_entry_destroy, NULL);
-	g_slist_free(ussd->ss_control_list);
+	g_slist_free_full(ussd->ss_control_list, ssc_entry_destroy);
 	ussd->ss_control_list = NULL;
 
-	g_slist_foreach(ussd->ss_passwd_list, (GFunc) ssc_entry_destroy, NULL);
-	g_slist_free(ussd->ss_passwd_list);
+	g_slist_free_full(ussd->ss_passwd_list, ssc_entry_destroy);
 	ussd->ss_passwd_list = NULL;
 
 	ofono_modem_remove_interface(modem,
